-- testdiff.vhd
-- Version 20090101 by whygee@f-cpu.org
-- Version 20090813 : changed port names,
--  added signed comparison, changed std_logic to std_ulogic

library ieee;
use ieee.std_logic_1164.all;

entity testdiff is
  generic (
  	TestNr     : integer := 1;  -- the test you want to run
    YASEP_SIZE : integer := 16;
    PIPELINED  : integer := 1   -- pipelined when 1
  );
  port(
    SND, SI4 : in std_ulogic_vector(YASEP_SIZE-1 downto 0);
    clk, add_EN, Addsub, ROP2_S1, ROP2_S2,
    ROP2_X2, ASU_ROP2, compare_signed, MulMux : in std_ulogic; -- ROP2 functions
    AuxLo, AuxHi : in std_ulogic_vector(11 downto 0);
    diff : out std_ulogic_vector(YASEP_SIZE downto 0));
end testdiff;

architecture toplevel of testdiff is
  signal SumDUT, SumDUTr, ROP2_OUTr : std_ulogic_vector(YASEP_SIZE downto 0);
  signal inA, inB, ROP2_OUTt, ROP2_OUTref : std_ulogic_vector(YASEP_SIZE-1 downto 0);
  signal AuxLoR, AuxHiR : std_ulogic_vector(11 downto 0);
  signal add_ENr, Addsubt, R2Ct, Addsubr, R2Cr, ROP2_S1r,
      ROP2_S2r, ROP2_X2r, compare_signed_r, MulMuxR : std_ulogic;
begin

----------------------------------------------------------------
-- This test indicates if "ActelProASIC3" is "equivalent" to "behaviour".
-- The result of synthesis should be "O"
----------------------------------------------------------------
test0 : if TestNr=0 generate

--    assert (PIPELINED=0)
--      report "PIPELINED must be 0 for test#0"
--      severity failure;

  inA <= (inA'range=>'0'); -- inA, inB and other pipeline gates are unused
  inB <= (inB'range=>'0'); --  because it's not pipelined.
  rop2_outr <= (rop2_outr'range=>'0');
  compare_signed_r <= compare_signed and (not MulMux);
  AuxLoR <= (AuxLoR'range=>'0');
  AuxHiR <= (AuxHiR'range=>'0');
  add_ENr  <= '0';
    -- force addsub to 0 when testing multiply
  Addsubt  <= Addsub and (not MulMux);
  Addsubr  <= '0';
  rop2_s1r <= '0';
  rop2_s2r <= '0';
  rop2_x2r <= '0';
  MulMuxR  <= '0';

  reference:
    entity work.ASUROP2_16(behaviour)
    generic map(
      PIPELINED  => 0,
      YASEP_SIZE => YASEP_SIZE
    )
    port map(
      SI4     => SI4,
      SND     => SND,
      Addsub  => Addsubt,
      clk     => clk,
      add_EN  => add_EN,
      add_ENr => add_EN,
      compare_signed => compare_signed_r,
      ROP2_S1 => ROP2_S1,
      ROP2_S2 => ROP2_S2,
      ROP2_X2 => ROP2_X2,
      MulMux  => MulMux,
      AuxLo   => AuxLo,
      AuxHi   => AuxHi,

      ROP2_OUT => ROP2_OUTref,
      ROP2_combine => R2Cr,
      Sum    => SumDUTr
    );

  ProASIC:
    entity work.ASUROP2_16(ActelProASIC3)
    generic map(
      PIPELINED  => 0,
      YASEP_SIZE => YASEP_SIZE
    )
    port map(
      SI4     => SI4,
      SND     => SND,
      Addsub  => Addsubt,
      clk     => clk,
      add_EN  => add_EN,
      add_ENr => add_EN,
      compare_signed => compare_signed,
      ROP2_S1 => ROP2_S1,
      ROP2_S2 => ROP2_S2,
      ROP2_X2 => ROP2_X2,
      MulMux  => MulMux,
      AuxLo   => AuxLo,
      AuxHi   => AuxHi,

      ROP2_OUT => ROP2_OUTt,
      ROP2_combine => R2Ct,
      Sum    => SumDUT
    );

    diff <= SumDUT xor SumDUTr 
     xor (ROP2_OUTref & R2Cr) xor (ROP2_OUTt & R2Ct);


  end generate;


----------------------------------------------------------------
-- This test indicates how fast ASUROP2_16 can be clocked.
----------------------------------------------------------------
test1 : if TestNr=1 generate

  -- remove warnings
  ROP2_OUTref <= (ROP2_OUTref'range=>'0');
  Addsubr <= '0';
  R2Cr <= '0';

  speed:
    entity work.ASUROP2_16(ActelProASIC3)
    generic map(
      PIPELINED  => PIPELINED,
      YASEP_SIZE => YASEP_SIZE
    )
    port map(
      SI4  => inA,
      SND  => inB,
      Addsub => Addsubt,
      clk    => clk,
      add_EN => add_ENr,
      add_ENr => add_ENr,
      compare_signed => compare_signed_r,
      ROP2_S1 => ROP2_S1r,
      ROP2_S2 => ROP2_S2r,
      ROP2_X2 => ROP2_X2r,
      MulMux  => MulMuxR,
      AuxLo   => AuxLoR,
      AuxHi   => AuxHiR,

      ROP2_OUT => ROP2_OUTt,
      ROP2_combine => R2Ct,
      Sum    => SumDUT
    );

    -- pipeline gates
    latche : process(clk)
    begin
      if (clk'event and clk='1') then
        inA  <= SI4;
        inB  <= SND;
        add_ENr <= add_EN;
        Addsubt <= Addsub;
        ROP2_S1r <= ROP2_S1;
        ROP2_S2r <= ROP2_S2;
        ROP2_X2r <= ROP2_X2;
        compare_signed_r <= compare_signed;
        MulMuxR <= MulMux;
        AuxLoR  <= AuxLo;
        AuxHiR  <= AuxHi;

        ROP2_OUTr <= R2Ct & ROP2_OUTt;
        SumDUTr <= SumDUT;
      end if;
    end process;

    -- some output
    diff <= SumDUTr when ASU_ROP2='0' else ROP2_OUTr;

  end generate;

end toplevel;
