-- RegSet16.vhd
-- version 2009-05-25
-- version 2009-06-03
-- version 2009-08-14

-- This ProASIC3 version is implemented as 2 independent blocks
-- of 256 words with 16-bit width,
-- Each block has simultaneous read and write ports.
-- There is 1 common write port and 2 independend read ports.
-- There are 16 banks of 16 registers.
-- It's not yet "optimised for power/toggle reduction".

library ieee;
use ieee.std_logic_1164.all;

library proasic3;
use proasic3.all;

entity RegSet16 is
  port (
    Clk, WrEn, -- Low=Write
    RdEn1, RdEn2 : in std_ulogic; -- active low
    AdrWr, AdrRd1, AdrRd2, BankID : in std_ulogic_vector(3 downto 0);
    DataIn: in std_ulogic_vector(15 downto 0);

    DataOut1, DataOut2 : out std_ulogic_vector(15 downto 0)
  );
end RegSet16;

architecture A3P of RegSet16 is

  component RAM512X18
    generic (MEMORYFILE:string := "");
    port(
      RADDR8, RADDR7, RADDR6, RADDR5, RADDR4, RADDR3, RADDR2, RADDR1, RADDR0,
      WADDR8, WADDR7, WADDR6, WADDR5, WADDR4, WADDR3, WADDR2, WADDR1, WADDR0,
      WD17, WD16, WD15, WD14, WD13, WD12, WD11, WD10,
      WD9, WD8, WD7, WD6, WD5, WD4, WD3, WD2, WD1, WD0,
      RW0, RW1, WW0, WW1, PIPE, REN, WEN, RCLK, WCLK, RESET : in std_logic := 'U';
      RD17, RD16, RD15, RD14, RD13, RD12, RD11, RD10,
      RD9, RD8, RD7, RD6, RD5, RD4, RD3, RD2, RD1, RD0 : out std_logic);
  end component;

begin

  RegSet1 : RAM512X18
  port map(
    RADDR8 => '0',
    RADDR7 => BankID(3),
    RADDR6 => BankID(2),
    RADDR5 => BankID(1),
    RADDR4 => BankID(0),
    RADDR3 => AdrWr(3),
    RADDR2 => AdrWr(2),
    RADDR1 => AdrWr(1),
    RADDR0 => AdrWr(0),
    
    WADDR8 => '0',
    WADDR7 => BankID(3),
    WADDR6 => BankID(2),
    WADDR5 => BankID(1),
    WADDR4 => BankID(0),
    WADDR3 => AdrRd1(3),
    WADDR2 => AdrRd1(2),
    WADDR1 => AdrRd1(1),
    WADDR0 => AdrRd1(0),

    WD17 => '0',
    WD16 => '0',
    WD15 => DataIn(15),
    WD14 => DataIn(14),
    WD13 => DataIn(13),
    WD12 => DataIn(12),
    WD11 => DataIn(11),
    WD10 => DataIn(10),
    WD9  => DataIn(9),
    WD8  => DataIn(8),
    WD7  => DataIn(7),
    WD6  => DataIn(6),
    WD5  => DataIn(5),
    WD4  => DataIn(4),
    WD3  => DataIn(3),
    WD2  => DataIn(2),
    WD1  => DataIn(1),
    WD0  => DataIn(0),

    RW0 => '0', RW1 => '1',
    WW0 => '0', WW1 => '1', PIPE => '0',
    REN => RdEn1, WEN => WrEn,
    RCLK => Clk, WCLK => Clk,
    RESET => '1',

    RD17 => open,
    RD16 => open,
    RD15 => DataOut1(15),
    RD14 => DataOut1(14),
    RD13 => DataOut1(13),
    RD12 => DataOut1(12),
    RD11 => DataOut1(11),
    RD10 => DataOut1(10),
    RD9 =>  DataOut1(9),
    RD8 =>  DataOut1(8),
    RD7 =>  DataOut1(7),
    RD6 =>  DataOut1(6),
    RD5 =>  DataOut1(5),
    RD4 =>  DataOut1(4),
    RD3 =>  DataOut1(3),
    RD2 =>  DataOut1(2),
    RD1 =>  DataOut1(1),
    RD0 =>  DataOut1(0)
  );

  RegSet2 : RAM512X18
  port map(
    RADDR8 => '0',
    RADDR7 => BankID(3),
    RADDR6 => BankID(2),
    RADDR5 => BankID(1),
    RADDR4 => BankID(0),
    RADDR3 => AdrWr(3),
    RADDR2 => AdrWr(2),
    RADDR1 => AdrWr(1),
    RADDR0 => AdrWr(0),
    
    WADDR8 => '0',
    WADDR7 => BankID(3),
    WADDR6 => BankID(2),
    WADDR5 => BankID(1),
    WADDR4 => BankID(0),
    WADDR3 => AdrRd2(3),
    WADDR2 => AdrRd2(2),
    WADDR1 => AdrRd2(1),
    WADDR0 => AdrRd2(0),


    WD17 => '0',
    WD16 => '0',
    WD15 => DataIn(15),
    WD14 => DataIn(14),
    WD13 => DataIn(13),
    WD12 => DataIn(12),
    WD11 => DataIn(11),
    WD10 => DataIn(10),
    WD9  => DataIn(9),
    WD8  => DataIn(8),
    WD7  => DataIn(7),
    WD6  => DataIn(6),
    WD5  => DataIn(5),
    WD4  => DataIn(4),
    WD3  => DataIn(3),
    WD2  => DataIn(2),
    WD1  => DataIn(1),
    WD0  => DataIn(0),

    RW0 => '0', RW1 => '1',
    WW0 => '0', WW1 => '1', PIPE => '0',
    REN => RdEn2, WEN => WrEn,
    RCLK => Clk, WCLK => Clk,
    RESET => '1',

    RD17 => open,
    RD16 => open,
    RD15 => DataOut2(15),
    RD14 => DataOut2(14),
    RD13 => DataOut2(13),
    RD12 => DataOut2(12),
    RD11 => DataOut2(11),
    RD10 => DataOut2(10),
    RD9 =>  DataOut2(9),
    RD8 =>  DataOut2(8),
    RD7 =>  DataOut2(7),
    RD6 =>  DataOut2(6),
    RD5 =>  DataOut2(5),
    RD4 =>  DataOut2(4),
    RD3 =>  DataOut2(3),
    RD2 =>  DataOut2(2),
    RD1 =>  DataOut2(1),
    RD0 =>  DataOut2(0)
  );

end A3P;
